@@ -1,9 +1,9 @@
-/*  Clzip - Data compressor based on the LZMA algorithm
-    Copyright (C) 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+/*  Clzip - LZMA lossless data compressor
+    Copyright (C) 2010-2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -15,7 +15,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 /*
-    Return values: 0 for a normal exit, 1 for environmental problems
+    Exit status: 0 for a normal exit, 1 for environmental problems
     (file not found, invalid flags, I/O errors, etc), 2 to indicate a
     corrupt or invalid input file, 3 for an internal consistency error
     (eg, bug) which caused clzip to panic.
@@ -52,10 +52,14 @@
 #endif
 
 #include "carg_parser.h"
-#include "clzip.h"
+#include "lzip.h"
 #include "decoder.h"
 #include "encoder.h"
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
@@ -63,15 +67,9 @@
 
 const char * const Program_name = "Clzip";
 const char * const program_name = "clzip";
-const char * const program_year = "2013";
+const char * const program_year = "2014";
 const char * invocation_name = 0;
 
-#ifdef O_BINARY
-const int o_binary = O_BINARY;
-#else
-const int o_binary = 0;
-#endif
-
 struct { const char * from; const char * to; } const known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
@@ -79,13 +77,11 @@ struct { const char * from; const char * to; } const known_extensions[] = {
 
 struct Lzma_options
   {
-  int dictionary_size;		/* 4KiB..512MiB */
-  int match_len_limit;		/* 5..273 */
+  int dictionary_size;		/* 4 KiB .. 512 MiB */
+  int match_len_limit;		/* 5 .. 273 */
   };
 
 enum Mode { m_compress, m_decompress, m_test };
-const unsigned long long max_member_size = 0x1000000000000000ULL;
-const unsigned long long max_volume_size = 0x7FFFFFFFFFFFFFFFULL;
 
 char * output_filename = 0;
 int outfd = -1;
@@ -98,7 +94,7 @@ bool delete_output_on_interrupt = false;
 
 static void show_help( void )
   {
-  printf( "%s - Data compressor based on the LZMA algorithm.\n", Program_name );
+  printf( "%s - LZMA lossless data compressor.\n", Program_name );
   printf( "\nUsage: %s [options] [files]\n", invocation_name );
   printf( "\nOptions:\n"
           "  -h, --help                     display this help and exit\n"
@@ -112,7 +108,7 @@ static void show_help( void )
           "  -m, --match-length=<bytes>     set match length limit in bytes [36]\n"
           "  -o, --output=<file>            if reading stdin, place the output into <file>\n"
           "  -q, --quiet                    suppress all messages\n"
-          "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8MiB]\n"
+          "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8 MiB]\n"
           "  -S, --volume-size=<bytes>      set volume size limit in bytes\n"
           "  -t, --test                     test compressed file integrity\n"
           "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
@@ -126,7 +122,12 @@ static void show_help( void )
           "The bidimensional parameter space of LZMA can't be mapped to a linear\n"
           "scale optimal for all files. If your files are large, very repetitive,\n"
           "etc, you may need to use the --match-length and --dictionary-size\n"
-          "options directly to achieve optimal performance.\n"
+          "options directly to achieve optimal performance. For example, -9m64\n"
+          "usually compresses executables more (and faster) than -9.\n"
+          "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
+          "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
+          "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
+          "caused clzip to panic.\n"
           "\nReport bugs to lzip-bug@nongnu.org\n"
           "Clzip home page: http://www.nongnu.org/lzip/clzip.html\n" );
   }
@@ -134,29 +135,28 @@ static void show_help( void )
 
 static void show_version( void )
   {
-  printf( "%s %s\n", Program_name, PROGVERSION );
+  printf( "%s %s\n", program_name, PROGVERSION );
   printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+  printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
           "This is free software: you are free to change and redistribute it.\n"
           "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
 
-void show_header( const File_header header )
+static void show_header( const unsigned dictionary_size )
   {
   const char * const prefix[8] =
     { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
   enum { factor = 1024 };
   const char * p = "";
   const char * np = "  ";
-  unsigned num = Fh_get_dictionary_size( header ), i;
+  unsigned num = dictionary_size, i;
   bool exact = ( num % factor == 0 );
 
   for( i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
     { num /= factor; if( num % factor != 0 ) exact = false;
       p = prefix[i]; np = ""; }
-  fprintf( stderr, "version %d, dictionary size %s%4u %sB.  ",
-           Fh_version( header ), np, num, p );
+  fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
   }
 
 
@@ -254,7 +254,7 @@ static int open_instream( const char * const name, struct stat * const in_statsp
     }
   else
     {
-    infd = open( name, O_RDONLY | o_binary );
+    infd = open( name, O_RDONLY | O_BINARY );
     if( infd < 0 )
       {
       if( verbosity >= 0 )
@@ -268,12 +268,13 @@ static int open_instream( const char * const name, struct stat * const in_statsp
       const bool can_read = ( i == 0 &&
                               ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                                 S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
-      if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
+      const bool no_ofile = to_stdout || program_mode == m_test;
+      if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
         {
         if( verbosity >= 0 )
           fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
                    program_name, name,
-                   ( can_read && !to_stdout ) ?
+                   ( can_read && !no_ofile ) ?
                    " and '--stdout' was not specified" : "" );
         close( infd );
         infd = -1;
@@ -334,7 +335,7 @@ static void set_d_outname( const char * const name, const int i )
 
 static bool open_outstream( const bool force )
   {
-  int flags = O_CREAT | O_WRONLY | o_binary;
+  int flags = O_CREAT | O_WRONLY | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename, flags, outfd_mode );
@@ -390,10 +391,14 @@ static void close_and_set_permissions( const struct stat * const in_statsp )
   bool warning = false;
   if( in_statsp )
     {
+    const mode_t mode = in_statsp->st_mode;
     /* fchown will in many cases return with EPERM, which can be safely ignored. */
-    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
-          errno != EPERM ) ||
-        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
+    if( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) == 0 )
+      { if( fchmod( outfd, mode ) != 0 ) warning = true; }
+    else
+      if( errno != EPERM ||
+          fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
+        warning = true;
     }
   if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
   outfd = -1;
@@ -431,6 +436,8 @@ static int compress( const unsigned long long member_size,
                      const int infd, struct Pretty_print * const pp,
                      const struct stat * const in_statsp )
   {
+  const unsigned long long cfile_size =
+    (in_statsp && S_ISREG( in_statsp->st_mode )) ? in_statsp->st_size / 100 : 0;
   unsigned long long in_size = 0, out_size = 0, partial_volume_size = 0;
   int retval = 0;
   struct Matchfinder matchfinder;
@@ -441,12 +448,12 @@ static int compress( const unsigned long long member_size,
   if( !Fh_set_dictionary_size( header, encoder_options->dictionary_size ) ||
       encoder_options->match_len_limit < min_match_len_limit ||
       encoder_options->match_len_limit > max_match_len )
-    internal_error( "invalid argument to encoder" );
+    internal_error( "invalid argument to encoder." );
 
   if( !Mf_init( &matchfinder, Fh_get_dictionary_size( header ),
                 encoder_options->match_len_limit, infd ) )
     {
-    Pp_show_msg( pp, "Not enough memory. Try a smaller dictionary size" );
+    Pp_show_msg( pp, "Not enough memory. Try a smaller dictionary size." );
     return 1;
     }
   Fh_set_dictionary_size( header, matchfinder.dictionary_size );
@@ -454,22 +461,24 @@ static int compress( const unsigned long long member_size,
   while( true )			/* encode one member per iteration */
     {
     struct LZ_encoder encoder;
-    const unsigned long long size = ( ( volume_size > 0 ) ?
-      min( member_size, volume_size - partial_volume_size ) : member_size );
+    const unsigned long long size = ( volume_size > 0 ) ?
+      min( member_size, volume_size - partial_volume_size ) : member_size;
     if( !LZe_init( &encoder, &matchfinder, header, outfd ) )
       {
       show_error( "Not enough memory. Try a smaller dictionary size.", 0, false );
       cleanup_and_fail( 1 );
       }
+    if( verbosity >= 2 )
+      show_progress( in_size, &matchfinder, pp, cfile_size );	/* init */
     if( !LZe_encode_member( &encoder, size ) )
-      { Pp_show_msg( pp, "Encoder error" ); retval = 1; break; }
+      { Pp_show_msg( pp, "Encoder error." ); retval = 1; break; }
     in_size += Mf_data_position( &matchfinder );
-    out_size += Re_member_position( &encoder.range_encoder );
+    out_size += Re_member_position( &encoder.renc );
     LZe_free( &encoder );
     if( Mf_finished( &matchfinder ) ) break;
     if( volume_size > 0 )
       {
-      partial_volume_size += Re_member_position( &encoder.range_encoder );
+      partial_volume_size += Re_member_position( &encoder.renc );
       if( partial_volume_size >= volume_size - min_dictionary_size )
         {
         partial_volume_size = 0;
@@ -477,7 +486,7 @@ static int compress( const unsigned long long member_size,
           {
           close_and_set_permissions( in_statsp );
           if( !next_filename() )
-            { Pp_show_msg( pp, "Too many volume files" ); retval = 1; break; }
+            { Pp_show_msg( pp, "Too many volume files." ); retval = 1; break; }
           if( !open_outstream( true ) ) { retval = 1; break; }
           delete_output_on_interrupt = true;
           }
@@ -512,13 +521,14 @@ static int decompress( const int infd, struct Pretty_print * const pp,
   bool first_member;
   if( !Rd_init( &rdec, infd ) )
     {
-    show_error( "Not enough memory. Find a machine with more memory.", 0, false );
+    show_error( "Not enough memory.", 0, false );
     cleanup_and_fail( 1 );
     }
 
   for( first_member = true; ; first_member = false )
     {
     int result;
+    unsigned dictionary_size;
     File_header header;
     struct LZ_decoder decoder;
     Rd_reset_member_position( &rdec );
@@ -526,13 +536,14 @@ static int decompress( const int infd, struct Pretty_print * const pp,
     if( Rd_finished( &rdec ) )			/* End Of File */
       {
       if( first_member )
-        { Pp_show_msg( pp, "Error reading member header" ); retval = 1; }
+        { Pp_show_msg( pp, "File ends unexpectedly at member header." );
+          retval = 2; }
       break;
       }
     if( !Fh_verify_magic( header ) )
       {
       if( !first_member ) break;		/* trailing garbage */
-      Pp_show_msg( pp, "Bad magic number (file not in lzip format)" );
+      Pp_show_msg( pp, "Bad magic number (file not in lzip format)." );
       retval = 2; break;
       }
     if( !Fh_verify_version( header ) )
@@ -543,17 +554,19 @@ static int decompress( const int infd, struct Pretty_print * const pp,
                    Fh_version( header ) ); }
       retval = 2; break;
       }
-    if( Fh_get_dictionary_size( header ) < min_dictionary_size ||
-        Fh_get_dictionary_size( header ) > max_dictionary_size )
-      { Pp_show_msg( pp, "Invalid dictionary size in member header" );
+    dictionary_size = Fh_get_dictionary_size( header );
+    if( dictionary_size < min_dictionary_size ||
+        dictionary_size > max_dictionary_size )
+      { Pp_show_msg( pp, "Invalid dictionary size in member header." );
         retval = 2; break; }
 
     if( verbosity >= 2 || ( verbosity == 1 && first_member ) )
-      { Pp_show_msg( pp, 0 ); if( verbosity >= 2 ) show_header( header ); }
+      { Pp_show_msg( pp, 0 );
+        if( verbosity >= 3 ) show_header( dictionary_size ); }
 
-    if( !LZd_init( &decoder, header, &rdec, outfd ) )
+    if( !LZd_init( &decoder, &rdec, dictionary_size, outfd ) )
       {
-      show_error( "Not enough memory. Find a machine with more memory.", 0, false );
+      show_error( "Not enough memory.", 0, false );
       cleanup_and_fail( 1 );
       }
     result = LZd_decode_member( &decoder, pp );
@@ -565,21 +578,19 @@ static int decompress( const int infd, struct Pretty_print * const pp,
         {
         Pp_show_msg( pp, 0 );
         if( result == 2 )
-          fprintf( stderr, "File ends unexpectedly at pos %llu\n",
+          fprintf( stderr, "File ends unexpectedly at pos %llu.\n",
                    partial_file_pos );
         else
-          fprintf( stderr, "Decoder error at pos %llu\n", partial_file_pos );
+          fprintf( stderr, "Decoder error at pos %llu.\n", partial_file_pos );
         }
       retval = 2; break;
       }
     if( verbosity >= 2 )
-      { if( testing ) fprintf( stderr, "ok\n" );
-        else fprintf( stderr, "done\n" ); Pp_reset( pp ); }
+      { fprintf( stderr, testing ? "ok\n" : "done\n" ); Pp_reset( pp ); }
     }
   Rd_free( &rdec );
   if( verbosity == 1 && retval == 0 )
-    { if( testing ) fprintf( stderr, "ok\n" );
-      else fprintf( stderr, "done\n" ); }
+    fprintf( stderr, testing ? "ok\n" : "done\n" );
   return retval;
   }
 
@@ -600,28 +611,6 @@ static void set_signals( void )
   }
 
 
-void Pp_init( struct Pretty_print * const pp, const char * const filenames[],
-              const int num_filenames, const int v )
-  {
-  unsigned stdin_name_len;
-  int i;
-  pp->name = 0;
-  pp->stdin_name = "(stdin)";
-  pp->longest_name = 0;
-  pp->verbosity = v;
-  pp->first_post = false;
-  stdin_name_len = strlen( pp->stdin_name );
-
-  for( i = 0; i < num_filenames; ++i )
-    {
-    const char * const s = filenames[i];
-    const int len = ( (strcmp( s, "-" ) == 0) ? stdin_name_len : strlen( s ) );
-    if( len > pp->longest_name ) pp->longest_name = len;
-    }
-  if( pp->longest_name == 0 ) pp->longest_name = stdin_name_len;
-  }
-
-
 void show_error( const char * const msg, const int errcode, const bool help )
   {
   if( verbosity >= 0 )
@@ -629,7 +618,7 @@ void show_error( const char * const msg, const int errcode, const bool help )
     if( msg && msg[0] )
       {
       fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
+      if( errcode > 0 ) fprintf( stderr, ": %s.", strerror( errcode ) );
       fprintf( stderr, "\n" );
       }
     if( help )
@@ -642,11 +631,34 @@ void show_error( const char * const msg, const int errcode, const bool help )
 void internal_error( const char * const msg )
   {
   if( verbosity >= 0 )
-    fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
+    fprintf( stderr, "%s: internal error: %s\n", program_name, msg );
   exit( 3 );
   }
 
 
+void show_progress( const unsigned long long partial_size,
+                    const struct Matchfinder * const m,
+                    struct Pretty_print * const p,
+                    const unsigned long long cfile_size )
+  {
+  static unsigned long long csize = 0;		/* file_size / 100 */
+  static unsigned long long psize = 0;
+  static const struct Matchfinder * mf = 0;
+  static struct Pretty_print * pp = 0;
+
+  if( m )					/* initialize static vars */
+    { csize = cfile_size; psize = partial_size; mf = m; pp = p; }
+  if( mf && pp )
+    {
+    const unsigned long long pos = psize + Mf_data_position( mf );
+    if( csize > 0 )
+      fprintf( stderr, "%4llu%%", pos / csize );
+    fprintf( stderr, "  %.1f MB\r", pos / 1000000.0 );
+    Pp_reset( pp ); Pp_show_msg( pp, 0 );	/* restore cursor position */
+    }
+  }
+
+
 int main( const int argc, const char * const argv[] )
   {
   /* Mapping from gzip/bzip2 style 1..9 compression modes
@@ -664,6 +676,8 @@ int main( const int argc, const char * const argv[] )
     { 3 << 23, 132 },		/* -8 */
     { 1 << 25, 273 } };		/* -9 */
   struct Lzma_options encoder_options = option_mapping[6];  /* default = "-6" */
+  const unsigned long long max_member_size = 0x0100000000000000ULL;
+  const unsigned long long max_volume_size = 0x4000000000000000ULL;
   unsigned long long member_size = max_member_size;
   unsigned long long volume_size = 0;
   const char * input_filename = "";
@@ -702,6 +716,7 @@ int main( const int argc, const char * const argv[] )
     { 'h', "help",            ap_no  },
     { 'k', "keep",            ap_no  },
     { 'm', "match-length",    ap_yes },
+    { 'n', "threads",         ap_yes },
     { 'o', "output",          ap_yes },
     { 'q', "quiet",           ap_no  },
     { 's', "dictionary-size", ap_yes },
@@ -717,7 +732,7 @@ int main( const int argc, const char * const argv[] )
   CRC32_init();
 
   if( !ap_init( &parser, argc, argv, options, 0 ) )
-    { show_error( "Memory exhausted.", 0, false ); return 1; }
+    { show_error( "Not enough memory.", 0, false ); return 1; }
   if( ap_error( &parser ) )				/* bad option */
     { show_error( ap_error( &parser ), 0, true ); return 1; }
 
@@ -728,8 +743,7 @@ int main( const int argc, const char * const argv[] )
     if( !code ) break;					/* no more options */
     switch( code )
       {
-      case '0':
-      case '1': case '2': case '3': case '4':
+      case '0': case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
                 encoder_options = option_mapping[code-'0']; break;
       case 'b': member_size = getnum( arg, 100000, max_member_size ); break;
@@ -741,6 +755,7 @@ int main( const int argc, const char * const argv[] )
       case 'k': keep_input_files = true; break;
       case 'm': encoder_options.match_len_limit =
                   getnum( arg, min_match_len_limit, max_match_len ); break;
+      case 'n': break;
       case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
       case 's': encoder_options.dictionary_size = get_dict_size( arg );
@@ -749,7 +764,7 @@ int main( const int argc, const char * const argv[] )
       case 't': program_mode = m_test; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
-      default : internal_error( "uncaught option" );
+      default : internal_error( "uncaught option." );
       }
     } /* end process options */
 
@@ -780,7 +795,7 @@ int main( const int argc, const char * const argv[] )
       ( filenames_given || default_output_filename[0] ) )
     set_signals();
 
-  Pp_init( &pp, filenames, num_filenames, verbosity );
+  Pp_init( &pp, filenames, num_filenames );
 
   output_filename = resize_buffer( output_filename, 1 );
   for( i = 0; i < num_filenames; ++i )
